'use strict';

Object.defineProperty(exports, '__esModule', {
    value: true
});

var _createClass = (function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ('value' in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; })();

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { 'default': obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError('Cannot call a class as a function'); } }

var _utilsLines = require('../utils/lines');

var _ElementAssert = require('./ElementAssert');

var _ElementAssert2 = _interopRequireDefault(_ElementAssert);

/**
 * Base class for Node, Token and Fragment.
 *
 * @name Element
 * @class
 * @abstract
 */

var Element = (function () {
    /**
     * @param {String} type
     * @param {Element[]} children
     */

    function Element(type, children) {
        _classCallCheck(this, Element);

        this._type = type;

        this._firstChild = null;
        this._lastChild = null;
        this._parentElement = null;
        this._nextSibling = null;
        this._previousSibling = null;
        this._childElements = null;

        if (children) {
            for (var i = 0; i < children.length; i++) {
                if (children[i]._parentElement) {
                    throw new Error('Cannot add element to several parents');
                }
            }
            this._setChildren(children);
        }
    }

    _createClass(Element, [{
        key: 'removeChild',

        // ==== Child Element Manipulation =================================================================================

        /**
         * Removes specified element from the element child list.
         *
         * @param {Element} element
         */
        value: function removeChild(element) {
            if (element._parentElement !== this) {
                throw new Error('The element to be removed is not a child of this element.');
            }

            var children = this._childElements.concat();
            var elementIndex = children.indexOf(element);
            children.splice(elementIndex, 1);

            this._setChildren(children);

            var ownerProgram = this.ownerProgram;
            if (ownerProgram) {
                ownerProgram._removeElementsFromSearchIndex([element]);
            }
        }
    }, {
        key: 'appendChild',

        /**
         * Appends specified element to the end of the child list.
         * Accepts multiple nodes using `Fragment`.
         *
         * @param {Element} newElement
         */
        value: function appendChild(newElement) {
            var children = undefined;
            var newElements = undefined;
            if (newElement.isFragment) {
                this._ensureCanAdoptFragment(newElement);
                newElements = newElement._childElements;
                children = this._childElements.concat(newElement._childElements);
            } else {
                if (newElement._parentElement) {
                    throw new Error('Remove element before adding again');
                }
                this._ensureCanAdopt(newElement);
                newElements = [newElement];
                children = this._childElements.concat(newElement);
            }

            this._setChildren(children);

            if (newElements) {
                var ownerProgram = this.ownerProgram;
                if (ownerProgram) {
                    ownerProgram._addElementsToSearchIndex(newElements);
                }
            }
        }
    }, {
        key: 'prependChild',

        /**
         * Prepends specified element to the beginning of the child list.
         * Accepts multiple nodes using `Fragment`.
         *
         * @param {Element} newElement
         */
        value: function prependChild(newElement) {
            var children = undefined;
            var newElements = undefined;
            if (newElement.isFragment) {
                this._ensureCanAdoptFragment(newElement);
                newElements = newElement._childElements;
                children = newElement._childElements.concat(this._childElements);
            } else {
                if (newElement._parentElement) {
                    throw new Error('Remove element before adding again');
                }
                this._ensureCanAdopt(newElement);
                newElements = [newElement];
                children = [newElement].concat(this._childElements);
            }

            this._setChildren(children);

            if (newElements) {
                var ownerProgram = this.ownerProgram;
                if (ownerProgram) {
                    ownerProgram._addElementsToSearchIndex(newElements);
                }
            }
        }
    }, {
        key: 'insertChildBefore',

        /**
         * Inserts specified element before specified reference child.
         * Accepts multiple nodes using `Fragment`.
         *
         * @param {Element} newElement
         * @param {Element} referenceChild
         */
        value: function insertChildBefore(newElement, referenceChild) {
            if (referenceChild._parentElement !== this) {
                throw new Error('Invalid reference child');
            }

            var index = this._childElements.indexOf(referenceChild);
            var childrenBefore = this._childElements.slice(0, index);
            var childrenAfter = this._childElements.slice(index);

            var children = undefined;
            var newElements = undefined;
            if (newElement.isFragment) {
                this._ensureCanAdoptFragment(newElement);
                newElements = newElement._childElements;
                children = childrenBefore.concat(newElement._childElements, childrenAfter);
            } else {
                if (newElement._parentElement) {
                    throw new Error('Remove element before adding again');
                }

                this._ensureCanAdopt(newElement);
                children = childrenBefore.concat(newElement, childrenAfter);
                newElements = [newElement];
            }

            this._setChildren(children);

            if (newElements) {
                var ownerProgram = this.ownerProgram;
                if (ownerProgram) {
                    ownerProgram._addElementsToSearchIndex(newElements);
                }
            }
        }
    }, {
        key: 'replaceChildren',

        /**
         * Replaces children from `firstRefChild` to `lastRefChild` with specified element.
         * Accepts multiple replacement nodes using `Fragment`.
         *
         * @param {Element} newElement
         * @param {Element} firstRefChild
         * @param {Element} lastRefChild
         */
        value: function replaceChildren(newElement, firstRefChild, lastRefChild) {
            if (!firstRefChild || firstRefChild._parentElement !== this) {
                throw new Error('Invalid first reference child');
            }

            if (!lastRefChild || lastRefChild._parentElement !== this) {
                throw new Error('Invalid last reference child');
            }

            var firstIndex = this._childElements.indexOf(firstRefChild);
            var lastIndex = this._childElements.indexOf(lastRefChild);

            if (firstIndex > lastIndex) {
                throw new Error('Invalid reference children order');
            }

            var childrenBefore = this._childElements.slice(0, firstIndex);
            var childrenAfter = this._childElements.slice(lastIndex + 1);
            var replacedChildren = this._childElements.slice(firstIndex, lastIndex + 1);

            var children = undefined;
            var newElements = undefined;
            if (newElement.isFragment) {
                this._ensureCanAdoptFragment(newElement);
                children = childrenBefore.concat(newElement._childElements, childrenAfter);
                newElements = newElement._childElements;
            } else {
                if (newElement._parentElement) {
                    throw new Error('Remove element before adding again');
                }

                this._ensureCanAdopt(newElement);
                children = childrenBefore.concat(newElement, childrenAfter);
                newElements = [newElement];
            }

            this._setChildren(children);

            for (var i = 0; i < replacedChildren.length; i++) {
                var replacedChild = replacedChildren[i];
                replacedChild._parentElement = null;
                replacedChild._previousSibling = null;
                replacedChild._nextSibling = null;
            }

            var ownerProgram = this.ownerProgram;
            if (ownerProgram) {
                ownerProgram._removeElementsFromSearchIndex(replacedChildren);
                if (newElements) {
                    ownerProgram._addElementsToSearchIndex(newElements);
                }
            }
        }
    }, {
        key: 'getChildrenBetween',

        /**
         * Returns array of child element from firstRefChild to lastRefChild (including reference children).
         *
         * @param {Element} firstRefChild
         * @param {Element} lastRefChild
         * @returns {Array}
         */
        value: function getChildrenBetween(firstRefChild, lastRefChild) {
            if (!firstRefChild || firstRefChild._parentElement !== this) {
                throw new Error('Invalid first reference child');
            }

            if (!lastRefChild || lastRefChild._parentElement !== this) {
                throw new Error('Invalid last reference child');
            }

            var firstIndex = this._childElements.indexOf(firstRefChild);
            var lastIndex = this._childElements.indexOf(lastRefChild);

            if (firstIndex > lastIndex) {
                throw new Error('Invalid reference children order');
            }

            return this._childElements.slice(firstIndex, lastIndex + 1);
        }
    }, {
        key: '_ensureCanAdopt',

        /**
         * Makes sure specified child is not already one of the parents of this element.
         * Throws error on failure.
         *
         * @param {Element} child
         * @private
         */
        value: function _ensureCanAdopt(child) {
            var element = this;
            while (element) {
                if (element === child) {
                    throw new Error('The new child element contains the parent.');
                }
                element = element._parentElement;
            }
        }
    }, {
        key: '_ensureCanAdoptFragment',

        /**
         * Calls _ensureCanAdopt for each fragment element.
         *
         * @param {Element} fragment
         * @private
         */
        value: function _ensureCanAdoptFragment(fragment) {
            var fragmentChild = fragment._firstChild;
            while (fragmentChild) {
                this._ensureCanAdopt(fragmentChild);
                fragmentChild = fragmentChild._nextSibling;
            }
        }
    }, {
        key: '_setChildren',

        /**
         * Assigns new children. Runs element syntax assertions.
         *
         * @param {Element[]} newChildren
         * @private
         */
        value: function _setChildren(newChildren) {
            this._acceptChildren(new _ElementAssert2['default'](newChildren));

            if (newChildren.length > 0) {
                var previousChild = newChildren[0];
                this._firstChild = previousChild;
                previousChild._parentElement = this;
                previousChild._previousSibling = null;
                if (newChildren.length > 1) {
                    var child = undefined;
                    for (var i = 1; i < newChildren.length; i++) {
                        child = newChildren[i];
                        child._parentElement = this;
                        child._previousSibling = previousChild;
                        previousChild._nextSibling = child;
                        previousChild = child;
                    }
                    child._nextSibling = null;
                    this._lastChild = child;
                } else {
                    this._lastChild = previousChild;
                }
            } else {
                this._firstChild = this._lastChild = null;
            }

            this._childElements = newChildren;
        }
    }, {
        key: '_acceptChildren',

        /**
         * Runs element syntax assertions. Should be implemented for every Node.
         *
         * @param {Element[]} children
         * @abstract
         */
        value: function _acceptChildren(children) {}
    }, {
        key: 'cloneElement',

        /**
         * Clones current Element structure.
         *
         * @returns {Element}
         */
        value: function cloneElement() {
            var clonedChildren = new Array(this._childElements.length);
            for (var i = 0; i < clonedChildren.length; i++) {
                clonedChildren[i] = this._childElements[i].cloneElement();
            }
            return new this.constructor(clonedChildren);
        }
    }, {
        key: 'type',

        /**
         * Element type.
         * Important: some Node types also exist in Tokens. Do not rely on type only, use isToken, isNode.
         *
         * @returns {String}
         */
        get: function get() {
            return this._type;
        }
    }, {
        key: 'isToken',

        /**
         * True if Element is a Token.
         *
         * @returns {Boolean}
         */
        get: function get() {
            return false;
        }
    }, {
        key: 'isWhitespace',

        /**
         * True if Element is a whitespace Token.
         *
         * @returns {Boolean}
         */
        get: function get() {
            return false;
        }
    }, {
        key: 'isComment',

        /**
         * True if Element is a comment Token.
         *
         * @returns {Boolean}
         */
        get: function get() {
            return false;
        }
    }, {
        key: 'isNonCodeToken',

        /**
         * True if Element is Comment or Whitespace.
         *
         * @returns {Boolean}
         */
        get: function get() {
            return this.isComment || this.isWhitespace;
        }
    }, {
        key: 'isNode',

        /**
         * True if Element is a Node.
         *
         * @returns {Boolean}
         */
        get: function get() {
            return false;
        }
    }, {
        key: 'isExpression',

        /**
         * True if Element is a Node which can be used as Expression.
         *
         * @returns {Boolean}
         */
        get: function get() {
            return false;
        }
    }, {
        key: 'isStatement',

        /**
         * True if Element is a Node which can be used as Statement.
         *
         * @returns {Boolean}
         */
        get: function get() {
            return false;
        }
    }, {
        key: 'isPattern',

        /**
         * True if Element is a Node which can be used as Pattern.
         *
         * @returns {Boolean}
         */
        get: function get() {
            return false;
        }
    }, {
        key: 'isAssignable',

        /**
         * True if Element is a Node which can be used as left part of Assignment.
         *
         * @returns {Boolean}
         */
        get: function get() {
            return false;
        }
    }, {
        key: 'isFragment',

        /**
         * True if Element is a Fragment.
         *
         * @returns {Boolean}
         */
        get: function get() {
            return false;
        }
    }, {
        key: 'parentElement',

        // ==== Traversing =================================================================================================

        /**
         * Parent element or null if parent element does not exist.
         *
         * @returns {Element|null}
         */
        get: function get() {
            return this._parentElement;
        }
    }, {
        key: 'ownerProgram',

        /**
         * Owner Program for this element or null if element does not have Program in its parent hierarchy.
         *
         * @returns {Element}
         */
        get: function get() {
            var element = this;
            while (element && !element._isProgram) {
                element = element._parentElement;
            }
            return element;
        }
    }, {
        key: 'nextSibling',

        /**
         * Next sibling within parent.
         * Null if element is the last child in parent or if element does not have parent.
         *
         * @returns {Element|null}
         */
        get: function get() {
            return this._nextSibling;
        }
    }, {
        key: 'previousSibling',

        /**
         * Previous sibling within parent.
         * Null if element is the first child in parent or if element does not have parent.
         *
         * @returns {Element|null}
         */
        get: function get() {
            return this._previousSibling;
        }
    }, {
        key: 'nextToken',

        /**
         * Next token. Null if token was not foung.
         *
         * @returns {Element|null}
         */
        get: function get() {
            if (this._nextSibling) {
                return this._nextSibling.firstToken;
            }

            if (this._parentElement) {
                return this._parentElement.nextToken;
            }

            return null;
        }
    }, {
        key: 'previousToken',

        /**
         * Previous token. Null if token was not foung.
         *
         * @returns {Element|null}
         */
        get: function get() {
            if (this._previousSibling) {
                return this._previousSibling.lastToken;
            }

            if (this._parentElement) {
                return this._parentElement.previousToken;
            }

            return null;
        }
    }, {
        key: 'nextCodeToken',

        /**
         * Next token (non-whitespace and non-comment). Null if token was not foung.
         *
         * @returns {Element|null}
         */
        get: function get() {
            var token = this.nextToken;
            while (token && (token.isWhitespace || token.isComment)) {
                token = token.nextToken;
            }
            return token;
        }
    }, {
        key: 'previousCodeToken',

        /**
         * Previous token (non-whitespace and non-comment). Null if token was not foung.
         *
         * @returns {Element|null}
         */
        get: function get() {
            var token = this.previousToken;
            while (token && (token.isWhitespace || token.isComment)) {
                token = token.previousToken;
            }
            return token;
        }
    }, {
        key: 'nextNonWhitespaceToken',

        /**
         * Next non-whitespace token. Null if token was not foung.
         *
         * @returns {Element|null}
         */
        get: function get() {
            var token = this.nextToken;
            while (token && token.isWhitespace) {
                token = token.nextToken;
            }
            return token;
        }
    }, {
        key: 'previousNonWhitespaceToken',

        /**
         * Previous non-whitespace token. Null if token was not foung.
         *
         * @returns {Element|null}
         */
        get: function get() {
            var token = this.previousToken;
            while (token && token.isWhitespace) {
                token = token.previousToken;
            }
            return token;
        }
    }, {
        key: 'firstToken',

        /**
         * First token inside element child tree.
         *
         * @returns {Element|null}
         */
        get: function get() {
            var element = this._firstChild;
            while (element && !element.isToken) {
                element = element._firstChild;
            }
            return element;
        }
    }, {
        key: 'lastToken',

        /**
         * Last token inside element child tree.
         *
         * @returns {Element|null}
         */
        get: function get() {
            var element = this._lastChild;
            while (element && !element.isToken) {
                element = element._lastChild;
            }
            return element;
        }
    }, {
        key: 'firstChild',

        /**
         * First child element. Null if element does not have children.
         *
         * @returns {Element|null}
         */
        get: function get() {
            return this._firstChild;
        }
    }, {
        key: 'lastChild',

        /**
         * Last child element. Null if element does not have children.
         *
         * @returns {Element|null}
         */
        get: function get() {
            return this._lastChild;
        }
    }, {
        key: 'childElements',

        /**
         * Direct children of the element.
         *
         * @returns {ElementList}
         */
        get: function get() {
            return this._childElements.concat();
        }
    }, {
        key: 'childCount',

        /**
         * Direct child count.
         *
         * @returns {Number}
         */
        get: function get() {
            return this._childElements.length;
        }
    }, {
        key: 'range',

        /**
         * Calculates and returns Element range.
         *
         * @returns {Number[]}
         */
        get: function get() {
            var counter = 0;

            var previous = this.previousToken;
            while (previous) {
                counter += previous.sourceCodeLength;
                previous = previous.previousToken;
            }

            return [counter, counter + this.sourceCodeLength];
        }
    }, {
        key: 'loc',

        /**
         * Calculates and returns Element loc.
         *
         * @returns {Object}
         */
        get: function get() {
            var prevToken = this.previousToken;
            var startColumn = 0;
            var startLine = 1;
            while (prevToken) {
                var lines = prevToken.sourceCodeLines;
                startColumn += lines[lines.length - 1].length;
                if (lines.length > 1) {
                    while (prevToken) {
                        startLine += prevToken.newlineCount;
                        prevToken = prevToken.previousToken;
                    }
                    break;
                }
                prevToken = prevToken.previousToken;
            }

            var elementLines = this.sourceCodeLines;
            var endLine = startLine + elementLines.length - 1;
            var endColumn = elementLines[elementLines.length - 1].length;

            if (startLine === endLine) {
                endColumn += startColumn;
            }

            return {
                start: {
                    line: startLine,
                    column: startColumn
                },
                end: {
                    line: endLine,
                    column: endColumn
                }
            };
        }
    }, {
        key: 'sourceCodeLength',

        // ==== Source Code ================================================================================================

        /**
         * Generated source code length.
         *
         * @returns {Number}
         */
        get: function get() {
            var length = 0;
            var child = this._firstChild;
            while (child) {
                length += child.sourceCodeLength;
                child = child._nextSibling;
            }
            return length;
        }
    }, {
        key: 'sourceCode',

        /**
         * Generated source code.
         *
         * @returns {String}
         */
        get: function get() {
            var code = '';
            var child = this._firstChild;
            while (child) {
                code += child.sourceCode;
                child = child._nextSibling;
            }
            return code;
        }
    }, {
        key: 'sourceCodeLines',

        /**
         * Generated source code lines.
         *
         * @returns {String[]}
         */
        get: function get() {
            return (0, _utilsLines.getLines)(this.sourceCode);
        }
    }, {
        key: 'newlineCount',

        /**
         * Generated source code line break count.
         *
         * @returns {Number}
         */
        get: function get() {
            var count = 0;
            var child = this._firstChild;
            while (child) {
                count += child.newlineCount;
                child = child._nextSibling;
            }
            return count;
        }
    }]);

    return Element;
})();

exports['default'] = Element;
module.exports = exports['default'];

// Override